/*
 * Copyright (c) 2016, Luca Fulchir<luca@fulchir.it>, All rights reserved.
 *
 * This file is part of "libRaptorQ".
 *
 * libRaptorQ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * libRaptorQ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * and a copy of the GNU Lesser General Public License
 * along with libRaptorQ.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "RaptorQ/v1/common.hpp"
#include "RaptorQ/v1/Operation.hpp"
#include <algorithm>
#include <atomic>
#include <cmath>
#include <limits>
#include <memory>
#include <mutex>
#include <vector>
#include <deque>
#include <utility>
#include <sstream>
#include <string>
#include <istream>
#include <fstream>


namespace RaptorQ__v1 {
namespace Impl {

using OPSVec = std::deque<Operation>;

std::vector<uint8_t> RAPTORQ_API ops_to_raw (const std::deque<Operation> &ops);
inline std::vector<uint8_t> RAPTORQ_API ops_to_raw (const std::deque<Operation>
                                                                        &ops)
{
    std::vector<uint8_t> ret;
    ret.reserve (ops.size() * 10);
    for (const auto &op : ops) {
        op.serialize(ret);
    }
    return ret;
}

OPSVec RAPTORQ_API raw_to_ops (const std::vector<uint8_t> &raw);
inline OPSVec RAPTORQ_API raw_to_ops (const std::vector<uint8_t> &raw)
{
    OPSVec ops;
    for(std::vector<uint8_t>::size_type i = 0; i < raw.size(); i++) {
        if (raw[i] == static_cast<uint8_t> (Operation::_t::SWAP)) {
            assert ((i + 4) < raw.size());
            uint16_t row_1 = Operation::deserialize_uint16(raw, i + 1);
            uint16_t row_2 = Operation::deserialize_uint16(raw, i + 3);
            i += 4;
            ops.emplace_back (Operation::_t::SWAP, row_1, row_2);
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::DIV)) {
            assert ((i + 3) < raw.size());
            uint16_t row_1 = Operation::deserialize_uint16(raw, i + 1);
            uint8_t scalar = raw[i + 3];
            i += 3;
            ops.emplace_back (Operation::_t::DIV, row_1, Octet(scalar));
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::ADD_MUL)) {
            assert ((i + 5) < raw.size());
            uint16_t row_1 = Operation::deserialize_uint16(raw, i + 1);
            uint16_t row_2 = Operation::deserialize_uint16(raw, i + 3);
            uint8_t scalar = raw[i + 5];
            i += 5;
            ops.emplace_back (Operation::_t::ADD_MUL, row_1, row_2,
                                                                Octet(scalar));
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::BLOCK)) {
            assert ((i + 8) < raw.size());
            uint16_t rows = Operation::deserialize_uint16(raw, i + 1);
            uint16_t cols = Operation::deserialize_uint16(raw, i + 3);
            uint32_t nnz  = Operation::deserialize_uint32(raw, i + 5);
            i += 8;

            assert ((i + nnz * 5) < raw.size());
            SparseMtx mtx = SparseMtx(rows, cols);
            std::vector<Eigen::Triplet<Octet>> triplets(nnz);
            for (uint32_t j = 0; j < nnz; j++) {
                uint16_t row = Operation::deserialize_uint16(raw, i + 1);
                uint16_t col = Operation::deserialize_uint16(raw, i + 3);
                triplets[j] = Eigen::Triplet<Octet>(row, col, Octet(raw[i + 5]));
                i += 5;
            }
            mtx.setFromTriplets(triplets.begin(), triplets.end());
            ops.emplace_back (Operation::_t::BLOCK, mtx);
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::REORDER)) {
            assert ((i + 4) < raw.size());
            uint32_t size  = Operation::deserialize_uint32(raw, i + 1);
            i += 4;

            assert ((i + size * 2) < raw.size());
            std::vector<uint16_t> vec;
            vec.reserve(size);
            for (uint32_t j = 0; j < size; j++) {
                uint16_t row = Operation::deserialize_uint16(raw, i + 1);
                vec.emplace_back(row);
                i += 2;
            }
            ops.emplace_back (Operation::_t::REORDER, vec);
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::NONE)) {
            ;
        } else {
            assert ("Invalid instruction!");
        }
    }
    return ops;
}


// NOTE: no locking!! it is done in the memory section.

template<typename User_Data>
class RAPTORQ_LOCAL OVS
{
public:
    OVS (const OVS&) = delete;
    OVS& operator= (const OVS&) = delete;
    OVS (OVS &&) = delete;
    OVS& operator= (OVS &&) = delete;

    static inline OVS *get()
    {
        // just drop the cache on exit time. Don't bother destroying
        // it correctly
        static OVS<User_Data> *instance = new OVS<User_Data>();
        return instance;
    }

    bool add (User_Data &&ops, int block_size);
    bool apply (Matrix &mtx, int block_size);
    bool available (int block_size);
    void load_cache (const std::string &directory);
    void store_cache (const std::string &directory);
private:
    OVS ();

    std::mutex biglock;
    std::vector<User_Data> store;
    int index_from_block_size(int block_size);
    std::vector<uint8_t> read_file(const std::string &filename);
    void write_file(const User_Data &ops, const std::string &filename);
};

template<typename User_Data>
OVS<User_Data>::OVS()
{
    store.resize(RQ_blocks_size);
}

template<typename User_Data>
int OVS<User_Data>::index_from_block_size (int block_size)
{
    for (int index = 0; index < RQ_blocks_size; index++) {
        if (RQ_blocks[index] == block_size) {
            return index;
        }
    }
    assert (block_size == 10 && "Invalid block_size");
    return 0;
}

template<typename User_Data>
std::vector<uint8_t> OVS<User_Data>::read_file(const std::string &filename)
{
    std::ifstream stream(filename.c_str(), std::ios::in | std::ios::binary);
    if (!stream.good()) {
        return std::vector<uint8_t>();
    }

    std::vector<uint8_t> vec((std::istreambuf_iterator<char>(stream)), std::istreambuf_iterator<char>());
    return vec;
}

template<typename User_Data>
void OVS<User_Data>::write_file(const User_Data &ops,
                                                const std::string &filename)
{
    auto vec = ops_to_raw(ops);
    std::ofstream stream(filename.c_str(),
                                        std::ios::out | std::ofstream::binary);
    if (stream.good()) {
        std::copy(vec.begin(), vec.end(),
                                    std::ostreambuf_iterator<char>(stream));

    }
}

template<typename User_Data>
void OVS<User_Data>::load_cache (const std::string &directory)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    for (int index = 0; index < RQ_blocks_size; index++) {
        std::stringstream ss;
        ss << directory << "/" << RQ_blocks[index] << ".ops";
        std::string s = ss.str();
        std::vector<uint8_t> raw = read_file(s);
        if (raw.size() != 0) {
            store[index] = raw_to_ops(raw);
        }
    }
}

template<typename User_Data>
void OVS<User_Data>::store_cache (const std::string &directory)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    std::cout << directory << std::endl;
    for (int index = 0; index < RQ_blocks_size; index++) {
        if (store[index].size() != 0) {
            std::stringstream ss;
            ss << directory << "/" << RQ_blocks[index] << ".ops";
            std::string s = ss.str();
            write_file(store[index], s);
        }
    }
}

template<typename User_Data>
bool OVS<User_Data>::available(int block_size)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    int index = index_from_block_size(block_size);
    return (store[index].size() != 0);
}

template<typename User_Data>
bool OVS<User_Data>::apply (Matrix &mtx, int block_size)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    int index = index_from_block_size(block_size);

    if (store[index].size() == 0)
        return false;

    User_Data &ops = store[index];
    for (const auto &op : ops)
        op.build_mtx (mtx);
    return true;
}

template<typename User_Data>
bool OVS<User_Data>::add (User_Data &&ops, int block_size)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    int index = index_from_block_size(block_size);

    if (store[index].size() == 0) {
        store[index] = std::move(ops);
    }

    return true;
}


} // namespace Impl
} // namespace RaptorQ__v1
